"""
Defines the registry of matchers and the standard set of matchers
"""

import re
import hamcrest as hc
from difflib import get_close_matches
from hamcrest.core.base_matcher import BaseMatcher
from hamcrest.library.collection.isdict_containingentries import IsDictContainingEntries
from hamcrest.library.collection.issequence_containing import IsSequenceContainingEvery
from hamcrest.core.helpers.wrap_matcher import wrap_matcher


__author__ = "Ivan -DrSlump- Montes"
__email__ = "drslump@pollinimini.net"
__license__ = "MIT"


# Words to ignore when looking up matchers
IGNORED_WORDS = ['should', 'to', 'be', 'a', 'an', 'is', 'the', 'as']

# Map of registered matchers as alias:callable
matchers = {}
# Map of normalized matcher aliases as normalized:alias
normalized = {}
# Help messages associated to matchers
helpmatchers = {}


# All textual representation types in Python 2/3
try:
    text_types = (basestring, str, unicode)  # python 2
except NameError:
    text_types = (str,)


class ContextManagerResult(object):
    """ When an expression is used in a `with` statement we capture the params
        in the __exit__ method of the expression context manager with this class,
        this allows to pass it to the matchers as the value to test, which is mostly
        useful for the raise/throw one.
    """
    def __init__(self, type_, value, trace):
        self.exc_type = type_
        self.exc_value = value
        self.trace = trace

    def __str__(self):
        """ Provide a suitable description of the exception for AnyOf/AllOf matchers """
        return repr(self.exc_value)


def register(matcher, *aliases):
    """ Register a matcher associated to one or more aliases. Each alias
        given is also normalized.
    """
    docstr = matcher.__doc__ if matcher.__doc__ is not None else ''
    helpmatchers[matcher] = docstr.strip()

    for alias in aliases:
        matchers[alias] = matcher
        # Map a normalized version of the alias
        norm = normalize(alias)
        normalized[norm] = alias
        # Map a version without snake case
        norm = norm.replace('_', '')
        normalized[norm] = alias


def unregister(matcher):
    """ Unregister a matcher (or alias) from the registry
    """
    # If it's a string handle it like an alias
    if isinstance(matcher, text_types) and matcher in matchers:
        matcher = matchers[matcher]

    # Find all aliases associated to the matcher
    aliases = [k for k, v in matchers.iteritems() if v == matcher]
    for alias in aliases:
        del matchers[alias]
        # Clean up the normalized versions
        norms = [k for k, v in normalized.iteritems() if v == alias]
        for norm in norms:
            del normalized[norm]

    # Remove help docstring
    if matcher in helpmatchers:
        del helpmatchers[matcher]

    return len(aliases) > 0


def normalize(alias):
    """ Normalizes an alias by removing adverbs defined in IGNORED_WORDS
    """
    # Convert from CamelCase to snake_case
    alias = re.sub(r'([a-z])([A-Z])', r'\1_\2', alias)
    # Ignore words
    words = alias.lower().split('_')
    words = filter(lambda w: w not in IGNORED_WORDS, words)
    return '_'.join(words)


def lookup(alias):
    """ Tries to find a matcher callable associated to the given alias. If
        an exact match does not exists it will try normalizing it and even
        removing underscores to find one.
    """

    if alias in matchers:
        return matchers[alias]
    else:
        norm = normalize(alias)
        if norm in normalized:
            alias = normalized[norm]
            return matchers[alias]

    # Check without snake case
    if -1 != alias.find('_'):
        norm = normalize(alias).replace('_', '')
        return lookup(norm)

    return None


def suggest(alias, max=3, cutoff=0.5):
    """ Suggest a list of aliases which are similar enough
    """

    aliases = matchers.keys()
    similar = get_close_matches(alias, aliases, n=max, cutoff=cutoff)

    return similar


def aliases():
    """ Obtain the list of aliases """
    return list(matchers.keys())


def alias_help(alias):
    """ Get help for the given alias """
    matcher = lookup(alias)
    if not matcher:
        return None
    return helpmatchers.get(matcher)


# Matchers should be defined with verbose aliases to allow the use of
# natural english where possible. When looking up a matcher common adverbs
# like 'to', 'be' or 'is' are ignored in the comparison.
register(hc.equal_to,
         'be_equal_to', 'be_equals_to', 'be_eql_to', 'be_eq_to')
register(hc.instance_of,
         'be_an_instance_of', 'be_a', 'be_an')
register(hc.same_instance,
         'be_the_same_instance_as', 'be_the_same_as', 'be')

register(hc.has_entry,
         'have_the_entry', 'contain_the_entry')
register(hc.has_entries,
         'have_the_entries', 'contain_the_entries')
register(hc.has_key,
         'have_the_key', 'contain_the_key')
register(hc.has_value,
         'have_the_value', 'contain_the_value')
register(hc.is_in,
         'be_in', 'be_into', 'be_contained_in')
register(hc.has_item,
         'have_the_item', 'contain_the_item')
register(hc.has_items,
         'have_the_items', 'contain_the_items')
register(hc.contains_inanyorder,
         'have_in_any_order', 'contain_in_any_order')
register(hc.contains,
         'have', 'contain')
register(hc.only_contains,
         'have_only', 'contain_only')
register(hc.close_to,
         'be_close_to')
register(hc.greater_than,
         'be_greater_than', 'be_greater', 'be_gt',
         'be_above',
         'be_more_than', 'be_more')
register(hc.greater_than_or_equal_to,
         'be_greater_than_or_equal_to', 'be_greater_or_equal', 'be_ge',
         'be_more_than_or_equal', 'be_more_or_equal',
         'be_at_least')
register(hc.less_than,
         'be_less_than', 'be_less', 'be_lt', 'be_below')
register(hc.less_than_or_equal_to,
         'be_less_than_or_equal_to', 'be_less_or_equal', 'be_le',
         'be_at_most')
register(hc.has_length,
         'have_length', 'have_len')
register(hc.has_property,
         'have_the_property', 'contain_the_property',
         'have_the_prop', 'contain_the_prop')
register(hc.has_string,
         'have_the_string', 'contain_the_string')
register(hc.equal_to_ignoring_case,
         'be_equal_to_ignoring_case')
register(hc.equal_to_ignoring_whitespace,
         'be_equal_to_ignoring_whitespace')
register(hc.contains_string,
         'substr', 'have_the_substr', 'contain_the_substr',
         'substring', 'have_the_substring', 'contain_the_substring')
register(hc.ends_with,
         'end_with')
register(hc.starts_with,
         'start_with', 'begin_with')
register(hc.anything,
         'be_anything', 'be_any')


class TypeMatcher(BaseMatcher):
    def _matches(self, item):
        return isinstance(item, self.__class__.types)

    def describe_to(self, description):
        description.append_text(self.__class__.expected)

    def describe_mismatch(self, item, description):
        description.append_text('was a %s ' % item.__class__.__name__)
        description.append_description_of(item)

    @classmethod
    def __call__(cls, *args, **kwargs):
        return cls()


class IsInteger(TypeMatcher):
    """ Check if the value is an integer """
    try:
        types = (int, long)
    except:
        types = (int)  # Python 3
    expected = 'an integer'


class IsFloat(TypeMatcher):
    """ Check if the value is a float """
    types = float
    expected = 'a float'


class IsComplex(TypeMatcher):
    """ Check if the value is a complex number """
    types = complex
    expected = 'a complex number'


class IsNumeric(TypeMatcher):
    """ Check if the value is a numeric type """
    try:
        types = (int, long, float, complex)  # python 2
    except NameError:
        types = (int, float, complex)

    expected = 'a numeric type'


class IsString(TypeMatcher):
    """ Check if the value is a string """
    types = text_types
    expected = 'a string'


class IsStr(TypeMatcher):
    """ Check if the value is a str """
    try:
        types = (basestring, str)  # python 2
    except NameError:
        types = str

    expected = 'a str'


class IsUnicode(TypeMatcher):
    """ Check if the value is an unicode string """
    try:
        types = unicode  # python 2
    except NameError:
        types = str

    expected = 'a unicode string'


class IsBinary(TypeMatcher):
    """ Check if value is a binary string """
    try:
        types = bytes  # python 3
    except NameError:
        types = str

    expected = 'a binary string'


class IsByteArray(TypeMatcher):
    """ Check if the value is a bytearray """
    types = bytearray
    expected = 'a bytearray'


class IsDict(TypeMatcher):
    """ Check if the value is a dict """
    types = dict
    expected = 'a dict'


class IsList(TypeMatcher):
    """ Check if the value is a list """
    types = list
    expected = 'a list'


class IsTuple(TypeMatcher):
    """ Check if the value is a tuple """
    types = tuple
    expected = 'a tuple'


class IsSet(TypeMatcher):
    """ Check if the value is a set """
    types = set
    expected = 'a set'


class IsFrozenSet(TypeMatcher):
    """ Check if the value is a frozenset """
    types = frozenset
    expected = 'a frozenset'


class IsBool(TypeMatcher):
    """ Check if the value is a bool """
    types = bool
    expected = 'a bool'


class IsFunction(TypeMatcher):
    """ Check if the value is a function """
    import types
    types = types.FunctionType
    expected = 'a function'


class IsGenerator(BaseMatcher):
    """ Checks if the value is a generator function """
    def _matches(self, item):
        import inspect
        return inspect.isgeneratorfunction(item)

    def describe_to(self, desc):
        desc.append_text('a generator function')


class IsClass(BaseMatcher):
    """ Check if the value is a class """
    def _matches(self, item):
        import inspect
        return inspect.isclass(item)

    def describe_to(self, desc):
        desc.append_text('a class')


register(IsInteger, 'be_an_integer', 'be_an_int')
register(IsFloat, 'be_a_float')
register(IsComplex, 'be_a_complex_number', 'be_a_complex')
register(IsNumeric, 'be_numeric')
register(IsString, 'be_a_string')
register(IsStr, 'be_a_str')
register(IsUnicode, 'be_an_unicode_string', 'be_an_unicode')
register(IsBinary, 'be_a_binary_string', 'be_a_binary')
register(IsByteArray, 'be_a_bytearray', 'be_a_byte_array')
register(IsDict, 'be_a_dictionary', 'be_a_dict')
register(IsList, 'be_a_list', 'be_an_array')
register(IsTuple, 'be_a_tuple')
register(IsSet, 'be_a_set')
register(IsFrozenSet, 'be_a_frozenset', 'be_a_frozen_set')
register(IsFunction, 'be_a_function', 'be_a_func')
register(IsBool, 'be_a_boolean', 'be_a_bool')
register(IsGenerator, 'be_a_generator')
register(IsClass, 'be_a_class')


class IsIterable(BaseMatcher):
    """ Checks if a value is iterable """
    def _matches(self, item):
        try:
            iter(item)
            return True
        except TypeError:
            return False

    def describe_to(self, description):
        description.append_text('an iterable value')

register(IsIterable, 'be_an_iterable')


class IsCallable(BaseMatcher):
    """ Check if a value is callable """
    def _matches(self, item):
        return hasattr(item, '__call__')

    def describe_to(self, desc):
        desc.append_text('a callable value')

register(IsCallable, 'be_callable', 'be_a_callable_value', 'can_be_called')


class IsNone(BaseMatcher):
    """ Check if a value is None """
    def _matches(self, item):
        return True if item is None else False

    def describe_to(self, desc):
        desc.append_text('a None')

register(IsNone, 'be_none', 'be_a_none_value')


class IsTrue(BaseMatcher):
    """ Check if a value is True """
    def _matches(self, item):
        return item is True

    def describe_to(self, desc):
        desc.append_text('a True')


class IsFalse(BaseMatcher):
    """ Check if a value is False """
    def _matches(self, item):
        return item is False

    def describe_to(self, desc):
        desc.append_text('a False')


class IsTruthy(BaseMatcher):
    """ Check if a value is truthy """
    def _matches(self, item):
        return True if item else False

    def describe_to(self, desc):
        desc.append_text('a truthy value')


class IsFalsy(BaseMatcher):
    """ Check if a value is falsy """
    def _matches(self, item):
        return True if not item else False

    def describe_to(self, desc):
        desc.append_text('a falsy value')

register(IsTrue, 'be_true')
register(IsFalse, 'be_false')
register(IsTruthy, 'be_a_truthy_value', 'be_truthy', 'be_ok')
register(IsFalsy, 'be_a_falsy_value', 'be_falsy', 'be_ko')


class IsEmpty(BaseMatcher):
    """ Check if a value is empty """
    def _matches(self, item):
        try:
            return not bool(len(item))
        except:
            return False

    def describe_to(self, desc):
        desc.append_text('an empty value')

    def describe_mismatch(self, item, desc):
        try:
            l = len(item)
            desc.append_text('has {0} elements'.format(l))
        except:
            desc.append_value(item)
            desc.append_text(' does not have a length')

register(IsEmpty, 'be_empty')


class RaisesError(BaseMatcher):
    """ Checks if calling the value raises an error """

    def __init__(self, expected=None, message=None, regex=None):
        self.expected = expected
        self.message = message
        self.regex = regex
        self.thrown = None

    def _matches(self, item):
        # support passing a context manager result
        if isinstance(item, ContextManagerResult):
            # Python <2.7 may provide a non exception value
            if isinstance(item.exc_value, Exception):
                self.thrown = item.exc_value
            elif item.exc_type is not None:
                try:
                    self.thrown = item.exc_type(*item.exc_value)
                except TypeError:
                    self.thrown = item.exc_type(item.exc_value)
            else:
                return False
        else:
            try:
                # support passing arguments by feeding a tuple instead of a callable
                if not callable(item) and getattr(item, '__getitem__', False):
                    item[0](*item[1:])
                else:
                    item()
                return False
            except:
                # This should capture any kind of raised value
                import sys
                self.thrown = sys.exc_info()[1]

        # Fail if we have defined an expected error type
        if self.expected and not isinstance(self.thrown, self.expected):
            return False

        # Apply message filters
        if self.message:
            return self.message == str(self.thrown)
        elif self.regex:
            return re.match(self.regex, str(self.thrown))

        return True

    def describe_to(self, desc):
        if self.thrown and self.message:
            desc.append_text('to raise an exception with message "%s"'
                             % self.message)
        elif self.thrown and self.regex:
            desc.append_text('to raise an exception matching /%s/'
                             % self.regex)
        else:
            desc.append_text('to raise an exception')
            if self.expected:
                try:
                    exps = map(lambda x: x.__name__, self.expected)
                except:
                    exps = [self.expected.__name__]
                desc.append_text(' of type <%s>' % '>, <'.join(exps))

    def describe_mismatch(self, item, desc):
        if self.thrown:
            desc.append_text('was ')
            desc.append_text('<%s>' % self.thrown.__class__.__name__)
            if self.message or self.regex:
                desc.append_text(' "%s"' % str(self.thrown))
        else:
            desc.append_text('no exception was raised')

register(RaisesError,
         'raise_an_error', 'raise_an_exception',
         'raises_an_error', 'raises_an_exception', 'raises', 'raise',
         'throw_an_error', 'throw_an_exception',
         'throws_an_error', 'throws_an_exception', 'throws', 'throw')



class Changes(BaseMatcher):
    """ Checks if calling a value changes something """

    def __init__(self, watch):
        self.watch = watch
        self.before = None
        self.after = None
        self.changed = False

    def _matches(self, item):
        # support passing arguments by feeding a tuple instead of a callable
        if not callable(item) and getattr(item, '__getitem__', False):
            func = item[0]
            params = item[1:]
        else:
            func = item
            params = []

        try:
            before = self.watcher()
        except TypeError:
            before = self.watcher

        # keep a snapshot of the value in case it's mutable
        from copy import deepcopy
        self.before = deepcopy(before)

        func(*params)

        try:
            self.after = self.watcher()
        except TypeError:
            self.after = self.watcher

        try:
            hc.assert_that(self.after, hc.equal_to(self.before))
            self.changed = False
        except AssertionError:
            self.changed = True

        return self.changed

    def describe_to(self, desc):
        desc.append_text('change something')

    def describe_mismatch(self, item, desc):
        # To support its proper use when negated we need to check if
        # the values actually changed or not
        if self.changed:
            desc.append_text('did change from ') \
                .append_value(self.before) \
                .append_text(' to ') \
                .append_value(self.after)
        else:
            desc.append_text('it didn\'t change from ') \
                .append_value(self.before)

register(Changes,
         'change', 'changes', 'modify', 'modifies')


class Callback(BaseMatcher):
    """ Checks against an user supplied callback. The callback
        can should return True to indicate a successful match or
        False to indicate an unsuccessful one.
    """

    def __init__(self, callback):
        self.callback = callback

    def _matches(self, item):
        self.error = None
        try:
            result = self.callback(item)
            # Returning an expectation assumes it's correct (no failure raised)
            from .expectation import Expectation
            return isinstance(result, Expectation) or bool(result)
        except AssertionError:
            # Just forward assertion failures
            raise
        except Exception as ex:
            self.error = str(ex)
            return False

    def describe_to(self, desc):
        desc.append_text('passses callback ')
        if (isinstance(self.callback, type(lambda: None))
                and self.callback.__name__ == '<lambda>'):
            desc.append_text(self.callback.__name__)
        else:
            desc.append_text('{0}'.format(self.callback))

    def describe_mismatch(self, item, desc):
        if self.error:
            desc.append_text('gave an exception "%s"' % self.error)
        else:
            desc.append_text('returned False')


register(Callback,
         'callback', 'pass', 'pass_callback')


class MockCalled(BaseMatcher):
    """ Support for checking if mocks where called from the Mock library
    """
    def _matches(self, item):
        if not hasattr(item, 'called'):
            raise Exception('Mock object does not have a <called> attribute')
        return item.called

    def describe_to(self, desc):
        desc.append_text('called')

    def describe_mismatch(self, item, desc):
        if item.called:
            desc.append_text('was called')
        else:
            desc.append_text('was not called')

register(MockCalled, 'called', 'invoked')


class RegexMatcher(BaseMatcher):
    """ Checks against a regular expression """

    def __init__(self, regex, flags=0):
        self.regex = regex

        if isinstance(flags, text_types):
            self.flags = 0
            for ch in flags.upper():
                self.flags |= getattr(re, ch)
        else:
            self.flags = flags

    def _matches(self, item):
        # Make sure we are matching against a string
        hc.assert_that(item, IsString())

        match = re.search(self.regex, item, self.flags)
        return match is not None

    def describe_to(self, desc):
        desc.append_text('matching ')
        desc.append_text('/{0}/'.format(self.regex))

register(RegexMatcher, 'match', 'match_regex', 'match_regexp', 'be_matched_by')


class IsObjectContainingEntries(IsDictContainingEntries):
    """Matches if object has the properties from a given dict whose values and
    keys satisfy a given matcher.

    Examples::
        :param inst: The instance or class.
        :param mismatch_description: The description in case of failure.

        have_properties({
            'prop1': should.eq('value1'),
            'prop2': should.eq('value2')
        })
    """
    hidden = set(['should_not', 'should_all', 'should_any', 'should', 'should_none'])

    def __init__(self, value_matchers=None, **kwargs):
        base_dict = {}

        if value_matchers is None:
            value_matchers = kwargs

        for key, value in value_matchers.items():
            base_dict[key] = wrap_matcher(value)
        super(IsObjectContainingEntries, self).__init__(base_dict)

    def matches(self, inst, mismatch_description=None):
        # Make sure we are matching against a dict
        try:
            keys = dir(inst)
            attributes = dict(
                (key, getattr(inst, key))
                for key in dir(inst)
                if not key.startswith('__')
                and key not in IsObjectContainingEntries.hidden
            )
        except Exception as ex:
            if mismatch_description:
                mismatch_description.append_text(
                    'unable to extract attributes from value: {0}'.format(ex))
            return False

        return super(IsObjectContainingEntries, self).matches(
            attributes, mismatch_description)

    def describe_to(self, desc):
        desc.append_text('a class as ')
        super(IsObjectContainingEntries, self).describe_to(desc)


register(IsObjectContainingEntries,
         'have_the_properties', 'contain_the_properties', 'have_the_attributes', 'contain_the_attributes',
         'have_props', 'contain_props', 'have_attrs', 'contain_attrs')


class IsSequenceContainingEveryInOrderSparse(IsSequenceContainingEvery):
    """
    Matches if a list contains every given element in the same order but with
    optional interleaved items.
    No optional elements matching the required ones are allowed.
    Mismatch description prioritizes missing items over wrong order.
    e.g. [1, 3, 4] IsSequenceContainingEveryInOrder [1, 4]
                   but NOT IsSequenceContainingEveryInOrder [4, 1]
                   and NOT IsSequenceContainingEveryInOrder [1, 4, 4]
    """

    def __init__(self, *element_matchers):
        delegates = [hc.has_item(e) for e in element_matchers]
        self.matcher_all = hc.all_of(*delegates)
        self.matcher_any = hc.any_of(*delegates)
        self.matcher_order = hc.contains(*element_matchers)
        self.order_seq = None

    def _matches(self, sequence):
        self.order_seq = None
        try:
            seq = list(sequence)
            if self.matcher_all.matches(seq):
                self.order_seq = [i for i in seq if self.matcher_any.matches([i])]
                return self.matcher_order.matches(self.order_seq)
            else:
                return False
        except TypeError:
            return False

    def describe_mismatch(self, item, mismatch_description):
        if self.order_seq is None:
            mismatch_description.append_text(' instead of a ')
            self.matcher_all.describe_mismatch(item, mismatch_description)
        else:
            self.matcher_order.describe_mismatch(self.order_seq, mismatch_description)
            mismatch_description.append_text(
                ' from candidate list elements: '
            ).append_description_of(self.order_seq).append_text(
                ' that satisfied those conditions from '
            ).append_description_of(item)

    def describe_to(self, description):
        self.matcher_all.describe_to(description)
        description.append_text(' in this specific order')


register(IsSequenceContainingEveryInOrderSparse,
         'contain_sparse', 'have_sparse', 'contain_sparse_in_order',
         'contain_in_order_sparse', 'have_every_in_order_sparse',
         'have_in_order_sparse', 'contain_every_in_order_sparse')



import unittest

from .dsl import DslTestCase
from .coordination import CoordinationTestCase
from .expect import ExpectTestCase
from .patch import PatchTestCase


def all_tests():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(DslTestCase))
    suite.addTest(unittest.makeSuite(CoordinationTestCase))
    suite.addTest(unittest.makeSuite(ExpectTestCase))
    suite.addTest(unittest.makeSuite(PatchTestCase))
    return suite



try:
    import unittest2 as unittest  # Python 2.6
except:
    import unittest
from pyshould import *
from pyshould import dsl
from pyshould import Expectation


class DslTestCase(unittest.TestCase):
    """ Simple tests for the exposed DSL symbols """

    def test_keywords(self):
        keywords = (
            'should', 'should_not', 'should_any', 'should_all',
            'should_none', 'should_either',
            'it', 'all_of', 'any_of', 'none_of'
        )

        for kw in keywords:
            self.assertTrue(kw in globals(),
                            msg='keyword "%s" not found as global' % kw)
            self.assertTrue(kw in dir(dsl),
                            msg='keyword "%s" not found in dsl' % kw)

    def test_expectations(self):
        self.assertIsInstance(should, Expectation)
        self.assertIsInstance(should_not, Expectation)
        self.assertIsInstance(should_any, Expectation)
        self.assertIsInstance(should_all, Expectation)
        self.assertIsInstance(should_none, Expectation)
        self.assertIsInstance(should_either, Expectation)

    def test_quantifiers(self):
        self.assertIsInstance(it(0), Expectation)
        self.assertIsInstance(all_of([]), Expectation)
        self.assertIsInstance(any_of([]), Expectation)
        self.assertIsInstance(none_of([]), Expectation)

    def test_should_property(self):
        a = it(10)
        it(10).should.eq(10)
        it(10).should_not.eq(20)

    def test_should(self):
        self.assertRaises(AssertionError, lambda: 0 | should.equal(1))

    def test_should_not(self):
        self.assertRaises(AssertionError, lambda: 0 | should_not.equal(0))

    def test_should_either(self):
        0 | should_either.equal(1).equal(0)

    def test_should_all(self):
        self.assertRaises(AssertionError, lambda: [0, 1] | should.equal(0))

    def test_should_any(self):
        self.assertRaises(AssertionError, lambda: [0, 1] | should.equal(2))

    def test_should_none(self):
        self.assertRaises(AssertionError, lambda: [0, 1] | should.equal(1))

    def test_it(self):
        self.assertRaises(AssertionError, lambda: it(0).equal(1))

    def test_all_of(self):
        self.assertRaises(AssertionError, lambda: all_of([1, 2]).equal(1))
        self.assertRaises(AssertionError, lambda: all_of((1, 2)).equal(1))
        self.assertRaises(AssertionError, lambda: all_of(1, 2).equal(1))

        all_of([1, 2]) | should.be_int

    def test_any_of(self):
        self.assertRaises(AssertionError, lambda: any_of([2, 2]).equal(1))
        self.assertRaises(AssertionError, lambda: any_of((2, 2)).equal(1))
        self.assertRaises(AssertionError, lambda: any_of(2, 2).equal(1))

    def test_none_of(self):
        self.assertRaises(AssertionError, lambda: none_of([2, 1]).equal(1))
        self.assertRaises(AssertionError, lambda: none_of((2, 1)).equal(1))
        self.assertRaises(AssertionError, lambda: none_of(2, 1).equal(1))

    def test_context_manager(self):
        def fail_not_thrown():
            with should.throw:
                pass
        self.assertRaises(AssertionError, fail_not_thrown)

        def fail_thrown():
            with should.not_throw:
                raise KeyError('foo')
        self.assertRaises(AssertionError, fail_thrown)

        def nested_expression_ok():
            with should.throw(TypeError):
                1 | should.equal(1)
        self.assertRaisesRegexp(AssertionError,
                                'TypeError', nested_expression_ok)

        def nested_expression_fail():
            with should.throw(TypeError):
                1 | should.equal(2)
        self.assertRaisesRegexp(AssertionError,
                                '<2>', nested_expression_fail)

        with should.throw(TypeError):
            raise TypeError('foo')

        with should.not_throw(TypeError):
            raise KeyError('foo')

        with should.throw:
            raise KeyError('foo')

        with should.not_throw:
            pass

        class FooError(Exception):
            def __init__(self, foo, msg=None):
                if not foo:
                    raise ValueError('foo is not defined')

        with should.throw(FooError):
            raise FooError(10)

    def test_expect_throw_non_empty_constructor_exception(self):
        with should.throw(NonEmptyConstructorException):
            raise NonEmptyConstructorException([])

    def test_expect_throw_str_exceptions(self):
        with should.throw(KeyError):
            raise KeyError()

        with should.throw(KeyError):
            object = {}
            object['non-existing-key']

    def test_match(self):
        "foo" | should.match('^\w+$')

    def test_match_flags(self):
        "FOO" | should.match('^[a-z]+$', 'i')

    def test_match_non_string(self):
        self.assertRaises(
            AssertionError,
            lambda: 10 | should.match('^\d+')
        )

    def test_callback_matcher(self):
        1 | should.pass_callback(lambda x: x == 1)

    def test_callback_matcher_nested_expectation(self):
        1 | should.pass_callback(lambda x: x | should.eq(1))

    def test_empty_matcher(self):
        [] | should.be_empty
        '' | should.be_empty
        {} | should.be_empty
        ['foo'] | should_not.be_empty
        'foo' | should_not.be_empty
        {'foo': 'foo'} | should_not.be_empty

    def test_matcher_composition(self):
        d = {'foo': 'bar'}
        d | should.have_value(should.eq('bar'))
        d | should.have_entry('foo', should.eq('baz').or_eq('bar'))
        d | should.have_key('foo').and_have_value(should.eq('bar'))

        self.assertRaises(
            AssertionError,
            lambda: d | should.have_entry('foo', should.eq('baz'))
        )

    def test_configuring_matchers(self):
        m1 = should.eq(1)
        m2 = should.eq(2)
        m1.resolve(1)
        m2.resolve(2)

        m1 = should.be_true
        m2 = should.be_false
        m1.resolve(True)
        m2.resolve(False)

    def test_nested_matchers(self):
        d = dict(a='A', b='B', c='C')
        d | should.have_entries({
            'a': should.be_str.and_not_empty,
            'b': should.be_str,
            'c': should_not.eq('Z')
        })

    def test_nested_matchers_failure(self):
        d = dict(a='A', b='B', c='C')
        try:
            d | should.have_entries({
                'a': should.be_str.and_not_empty,
                'z': should.be_str,
                'c': should_not.eq('Z')
            })
            raise RuntimeError('We should not reach this point')
        except AssertionError:
            pass

    def test_class_has_props(self):
        d = type("CommentForm", (object,),
                 {'text': 'textvalue', 'author': 'the_author'})

        d | should.have_properties({
            'text': 'textvalue',
            'author': 'the_author',
        })

    def test_object_has_props(self):
        Cls = type("CommentForm", (object,), {'text': '', 'author': ''})
        d = Cls()
        d.text = 'textvalue'
        d.author = 'the_author'

        d | should.have_properties({
            'text': 'textvalue',
            'author': 'the_author',
        })

    def test_object_has_dynamic_props(self):
        Cls = type("CommentForm", (object,), {})
        d = Cls()
        d.added_text = 'textvalue'
        d.added_author = 'the_author'

        d | should.have_attrs({
            'added_text': 'textvalue',
            'added_author': should.be_a_string(),
        })

    def test_object_has_dynamic_props_with_kwargs(self):
        Cls = type("CommentForm", (object,), {})
        d = Cls()
        d.added_text = 'textvalue'
        d.added_author = 'the_author'

        d | should.have_attrs(added_text='textvalue',
                              added_author=should.be_a_string())

    def test_object_has_dynamic_props_with_non_existing_kwargs(self):
        Cls = type("CommentForm", (object,), {})
        d = Cls()
        d.added_text = 'textvalue'
        d.added_author = 'the_author'
        with self.assertRaises(AssertionError):
            d | should.have_attrs(added_text='textvalue',
                                  added_author=should.be_a_string(),
                                  fake='foo')

    def test_has_props_non_class(self):
        d = "fail"

        with self.assertRaises(AssertionError):
            d | should.have_properties({
                'text': should.be_a_string(),
                'author': 'the_author',
            })

    def test_old_style_class_has_props(self):
        class CommentForm:
            text_field = "textvalue"
            author_field = "the_author"

        d = CommentForm()

        d | should.have_properties({
            'text_field': 'textvalue',
            'author_field': 'the_author',
        })

    def test_apply(self):
        import json
        d = '{"foo":"bar"}'

        d | should(json.loads).have_key('foo')

        should_json = should(json.loads)
        d | should_json.have_key('foo')
        '{"bar": 10}' | should_json.have_key('bar')

        try:
            d | should(json.loads).have_key('bar')
            raise RuntimeError('We should not reach this point')
        except AssertionError:
            pass

    def test_apply_error(self):
        import json
        d = '{malformed}'

        try:
            d | should(json.loads).be_anything
            raise RuntimeError('We should not reach this point')
        except AssertionError:
            pass

    def test_apply_all(self):
        import json
        d = ('{"bar": 10}', '{"bar": 20}')

        d | should_all(json.loads).have_key('bar')

        should_all_json = should_all(json.loads)
        d | should_all_json.have_key('bar')

    def test_equality(self):
        m = should.be_int.and_eq(1)
        self.assertEqual(1, m)

        self.assertRaises(
            AssertionError,
            lambda: self.assertEqual(m, 2)
        )

    def test_inequality(self):
        m = should.be_int.and_eq(2)
        self.assertNotEqual(1, m)

        self.assertRaises(
            AssertionError,
            lambda: self.assertNotEqual(m, 2)
        )

    def test_equality_exception(self):
        # under the hood the equality for `should` should trigger an error
        (10 == should) | should.be_False
        (10 == should.equal) | should.be_False
        (10 != should) | should.be_True
        (10 != should.equal) | should.be_True

    def test_mock(self):
        try:
            from unittest.mock import Mock, patch  # Python 3.3
        except:
            try:
                from mock import Mock, patch
            except:
                raise unittest.SkipTest('Mock library not available, skipping test')

        mock = Mock()
        mock(10, 1)

        mock | should.be_called
        self.assertRaises(
            AssertionError,
            lambda: mock | should.not_be_called
        )

        # Check it supports matchers in params
        mock.assert_called_with(should.any, should.less_than(3))
        self.assertRaises(
            AssertionError,
            lambda: mock.assert_called_with(should.any, should.greater_than(3))
        )

        # Check we can wrap it even when implementing __or__
        with patch.object(self, 'test_mock') as mock:
            it(mock) | should.not_be_called

    def test_patch_mockito(self):
        import warnings
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            from pyshould import patch_mockito
            w | should.have_len(1)
            w[0].category | should.be(DeprecationWarning)

    def test_compatibility_with_assert(self):

        assert 10 | should.be_less_than(20)

        try:
            assert 10 | should.be_greater_than(20)
            raise RuntimeError('Should not reach this point')
        except AssertionError:
            pass

    def test_repr_expr(self):
        expr = it(10)
        repr(expr) | should.be_a_string

        expr = it(10).should.be_an_int
        repr(expr) | should.be_a_string

        expr = it(10).should.eq(10)
        repr(expr) | should.be_a_string

    def test_repr_deferred_expr(self):
        repr(should.be_empty) | should.be_a_string

    def test_base_except(self):
        import sys
        with should.throw(SystemExit):
            sys.exit(66)

    def test_base_except_not_thrown(self):
        with should_not.throw(SystemExit):
            pass

    def test_wrap_value(self):
        it(10) | should.eq(10)

    def test_dumper(self):
        output = []
        mockdumper = dumper(reporter=output.append)

        data = {'foo': 'Foo', 'bar': 'Bar', 'baz': 'Baz'}
        data | should.eq({
            'foo': mockdumper(msg='this is foo'),
            'bar': mockdumper,
            'baz': mockdumper(should.start_with('B'), msg='BAZ!')
        })

        output.sort()
        output | should.eq([
            "'Bar'",
            "BAZ!: 'Baz'",
            "this is foo: 'Foo'",
        ])

    def test_contain_sparse_in_order(self):
        with self.assertRaises(AssertionError):
            [1, 4, 3, 3, 3, 6] | should.contain_sparse_in_order(
                should.eq(1), should.be_greater_than(7)
            )

        with self.assertRaises(AssertionError):
            [1, 4, 3, 3, 3, 8, 9] | should.contain_sparse_in_order(
                should.eq(1), should.be_greater_than(7)
            )

        with self.assertRaises(AssertionError):
            [9, 1] | should.contain_sparse_in_order(
                should.eq(1), should.be_greater_than(7)
            )

        with self.assertRaises(AssertionError):
            1 | should.contain_sparse_in_order(
                should.eq(1)
            )

        with self.assertRaises(AssertionError):
            [1] | should.contain_sparse_in_order(
                should.eq(1), should.be_greater_than(7)
            )

        [1, 4, 3, 3, 8, 6, 2] | should.contain_sparse_in_order(
            should.eq(1), should.be_greater_than(7)
        )

        [0, 1, 4, 3, 3, 8, 6, 2] | should.contain_sparse_in_order(
            should.eq(1), should.be_greater_than(7)
        )

        (i for i in range(9)) | should.contain_sparse_in_order(
            1, should.be_greater_than(7)
        )


class NonEmptyConstructorException(Exception):

    parameter = []

    @property
    def text(self):
        return self.parameter

    def __init__(self, parameter, message=None):
        if parameter is None:
            raise TypeError("parameter could not be None")

        self.parameter = parameter
        self.message = message

    def __str__(self):
        return unicode(self).encode('utf-8')

    def __unicode__(self):
        return 'Missing fields: %s' % (self.parameter)



__author__ = "Ivan -DrSlump- Montes"
__email__ = "drslump@pollinimini.net"
__license__ = "MIT"

import os


class Dumper(object):
    """ Dumps a value when the equality comparison is triggered
    """

    UNSET_VALUE = {}
    REPORTER = lambda x: os.sys.stdout.write(x + os.linesep)

    def __init__(self, value=UNSET_VALUE, msg=None, reporter=REPORTER):
        self.value = value
        self.msg = msg
        self.reporter = reporter

    def __call__(self, value=UNSET_VALUE, msg=None, reporter=None):
        """ Generate a new dumper based on the configuration 
        """
        return Dumper(value=value, msg=msg, reporter=reporter or self.reporter)

    def __eq__(self, other):
        if self.msg != None:
            self.reporter(self.msg + ': ' + repr(other))
        else:
            self.reporter(repr(other))

        if self.value is not Dumper.UNSET_VALUE:
            return self.value == other

        return True

    def __str__(self):
        return repr(self)

    def __repr__(self):
        if self.value is Dumper.UNSET_VALUE:
            return '<dumper>'
        return repr(self.value)



"""
Patches cpython's root object to expose should assertions as 
properties.

The code is heavily based on the one originally found in the
sure library (https://github.com/gabrielfalcao/sure), used
with permission (https://github.com/gabrielfalcao/sure/issues/50)
"""

import platform

__author__ = "Ivan -DrSlump- Montes"
__email__ = "drslump@pollinimini.net"
__license__ = "MIT"


is_cpython = (
    hasattr(platform, 'python_implementation')
    and platform.python_implementation().lower() == 'cpython')

if is_cpython:

    import ctypes

    DictProxyType = type(object.__dict__)

    Py_ssize_t = \
        hasattr(ctypes.pythonapi, 'Py_InitModule4_64') \
            and ctypes.c_int64 or ctypes.c_int

    class PyObject(ctypes.Structure):
        pass

    PyObject._fields_ = [
        ('ob_refcnt', Py_ssize_t),
        ('ob_type', ctypes.POINTER(PyObject)),
    ]

    class SlotsProxy(PyObject):
        _fields_ = [('dict', ctypes.POINTER(PyObject))]

    def patchable_builtin(klass):
        name = klass.__name__
        target = getattr(klass, '__dict__', name)

        if not isinstance(target, DictProxyType):
            return target

        proxy_dict = SlotsProxy.from_address(id(target))
        namespace = {}

        ctypes.pythonapi.PyDict_SetItem(
            ctypes.py_object(namespace),
            ctypes.py_object(name),
            proxy_dict.dict,
        )

        return namespace[name]

    try:
        import __builtin__ as builtins
    except ImportError:
        import builtins

    def make_prop(exp, is_none=False):
        return builtins.property(
            fget=lambda self: exp(self),
            fset=lambda self, other: None,
            fdel=lambda self, *args, **kwargs: None
        )

    from pyshould.expectation import Expectation, ExpectationNot, ExpectationAll, \
                                     ExpectationAny, ExpectationNone

    object_handler = patchable_builtin(object)
    object_handler['should'] = make_prop(Expectation)
    object_handler['should_not'] = make_prop(ExpectationNot)
    object_handler['should_all'] = make_prop(ExpectationAll)
    object_handler['should_any'] = make_prop(ExpectationAny)
    object_handler['should_none'] = make_prop(ExpectationNone)

    # None does not have a tp_dict associated to its PyObject, so this
    # is the only way we could make it work like we expected.
    none_handler = patchable_builtin(None.__class__)
    none_handler['should'] = Expectation(None)
    none_handler['should_not'] = ExpectationNot(None)
    none_handler['should_all'] = ExpectationAll(None)
    none_handler['should_any'] = ExpectationAny(None)
    none_handler['should_none'] = ExpectationNone(None)

else:
    from warnings import warn
    warn("pyshould's patch is only supported on cpython", DeprecationWarning)
    


"""
Prints the configured matchers when run with `python -m pyshould`
"""

def main():
    from pyshould.matchers import lookup, aliases, alias_help
    group = {}
    for alias in aliases():
        m = lookup(alias)
        if m not in group:
            group[m] = []
        group[m].append(alias)

    for items in group.values():
        print("{0}:\n\t{1}\n".format(
            ', '.join(items),
            alias_help(items[0]).replace("\n", "\n\t")
        ))


main()



try:
    import unittest2 as unittest  # Python 2.6
except:
    import unittest

import pyshould.patch


class PatchTestCase(unittest.TestCase):
    """ Simple tests for the patched object api """

    def setUp(self):
        import platform
        is_cpython = (
            hasattr(platform, 'python_implementation')
            and platform.python_implementation().lower() == 'cpython')

        if not is_cpython:
            raise unittest.SkipTest(
                'patch is only available under cpython, skipping test')

    def test_should(self):
        from pyshould.expectation import Expectation
        assert isinstance(self.should, Expectation)
        self.should.be_a(unittest.TestCase)

    def test_should_apply(self):
        import json
        '{"foo":"bar"}'.should(json.loads).have_key('foo')

    def test_should_int(self):
        (1).should.eq(1)
        (3).should.not_eq(2)

    def test_should_str(self):
        'foo'.should.eq('foo')
        'foo'.should.not_eq('bar')

    def test_should_none(self):
        None.should.be_none()

    def test_should_not(self):
        (1).should_not.eq(2)

    def test_should_not_none(self):
        None.should_not.eq(2)

    def test_should_all(self):
        [1, 2, 3].should_all.be_int()

    def test_should_any(self):
        [1, 2, 3].should_any.eq(2)

    def test_should_none(self):
        [1, 2, 3].should_none.eq(5)

    def test_should_all_none(self):
        try:
            None.should_all.be_int()
            raise RuntimeError('We should not reach this point')
        except AssertionError:
            pass

    def test_should_fail(self):
        try:
            'foo'.should.eq('bar')
            raise RuntimeError('We should not reach this point')
        except AssertionError:
            pass

    def test_should_fail_none(self):
        try:
            None.should.eq('bar')
            raise RuntimeError('We should not reach this point')
        except AssertionError:
            pass



"""
Defines the expectation class which is the basis for performing the assertions.
"""
import re
import hamcrest as hc

from .patched import IsNot
from .matchers import lookup, suggest, ContextManagerResult

__author__ = "Ivan -DrSlump- Montes"
__email__ = "drslump@pollinimini.net"
__license__ = "MIT"


class OPERATOR:
    """ Define the set of coordination operators assigning to them a weight
        to indicate their precedence. Higher values take precedence over
        lower ones.
    """
    NOT = 20
    AND = 10
    OR = 5
    BUT = 1


class Expectation(object):
    """ Represents an expectation allowing to configure it with matchers and
        finally resolving it.
    """

    _contexts = []

    def __init__(self, value=None, deferred=False, description=None, factory=False,
                 def_op=OPERATOR.AND, def_matcher='equal'):
        self.reset()
        self.value = value
        self.deferred = deferred
        self.factory = factory
        self.description = description
        self.def_op = def_op
        self.def_matcher = def_matcher
        self.transform = None

    def reset(self):
        """ Resets the state of the expression """
        self.expr = []
        self.matcher = None
        self.last_matcher = None
        self.description = None

    def clone(self):
        """ Clone this expression """
        from copy import copy
        clone = copy(self)
        clone.expr = copy(self.expr)
        clone.factory = False
        return clone

    def __ror__(self, lvalue):
        """ Evaluate against the left hand side of the OR (pipe) operator. Since in
            Python this operator has a fairly low precedence this method will usually
            be called once the whole right hand side of the expression has been evaluated.

            Note: We clone and return that clone instead of the self object because
                  resolving resets the expectation, when using a REPL it's nice to see
                  the expectation explanation after a successful one.
        """
        clone = self.clone()
        self.resolve(lvalue)
        return clone

    def __or__(self, rvalue):
        """ Allows the use case: it(value) | should.xxx
            Specially useful to wrap mocks which override the __or__ operator
        """
        if isinstance(rvalue, Expectation) and self.__class__ is Expectation:
            return rvalue.__ror__(self.value)

        return NotImplemented

    def resolve(self, value=None):
        """ Resolve the current expression against the supplied value """

        # If we still have an uninitialized matcher init it now
        if self.matcher:
            self._init_matcher()

        # Evaluate the current set of matchers forming the expression
        matcher = self.evaluate()

        try:
            value = self._transform(value)
            self._assertion(matcher, value)
        except AssertionError as ex:
            # By re-raising here the exception we reset the traceback
            raise ex
        finally:
            # Reset the state of the object so we can use it again
            if self.deferred:
                self.reset()

    def _assertion(self, matcher, value):
        """ Perform the actual assertion for the given matcher and value. Override
            this method to apply a special configuration when performing the assertion.
            If the assertion fails it should raise an AssertionError.
        """
        # To support the syntax `any_of(subject) | should ...` we check if the
        # value to check is an Expectation object and if it is we use the descriptor
        # protocol to bind the value's assertion logic to this expectation.
        if isinstance(value, Expectation):
            assertion = value._assertion.__get__(self, Expectation)
            assertion(matcher, value.value)
        else:
            hc.assert_that(value, matcher)

    def _transform(self, value):
        """ Applies any defined transformation to the given value
        """
        if self.transform:
            try:
                value = self.transform(value)
            except:
                import sys
                exc_type, exc_obj, exc_tb = sys.exc_info()
                raise AssertionError('Error applying transformation <{0}>: {2}: {3}'.format(
                    self.transform.__name__, value, exc_type.__name__, exc_obj))

        return value

    def evaluate(self):
        """ Converts the current expression into a single matcher, applying
            coordination operators to operands according to their binding rules
        """

        # Apply Shunting Yard algorithm to convert the infix expression
        # into Reverse Polish Notation. Since we have a very limited
        # set of operators and binding rules, the implementation becomes
        # really simple. The expression is formed of hamcrest matcher instances
        # and operators identifiers (ints).
        ops = []
        rpn = []
        for token in self.expr:
            if isinstance(token, int):
                while len(ops) and token <= ops[-1]:
                    rpn.append(ops.pop())
                ops.append(token)
            else:
                rpn.append(token)

        # Append the remaining operators
        while len(ops):
            rpn.append(ops.pop())

        # Walk the RPN expression to create AllOf/AnyOf matchers
        stack = []
        for token in rpn:
            if isinstance(token, int):
                # Handle the NOT case in a special way since it's unary
                if token == OPERATOR.NOT:
                    stack[-1] = IsNot(stack[-1])
                    continue

                # Our operators always need two operands
                if len(stack) < 2:
                    raise RuntimeError('Unable to build a valid expression. Not enough operands available.')

                # Check what kind of matcher we need to create
                if token == OPERATOR.OR:
                    matcher = hc.any_of(*stack[-2:])
                else:  # AND, BUT
                    matcher = hc.all_of(*stack[-2:])

                stack[-2:] = [matcher]
            else:
                stack.append(token)

        if len(stack) != 1:
            raise RuntimeError('Unable to build a valid expression. The RPN stack should have just one item.')

        matcher = stack.pop()

        # If a description has been given include it in the matcher
        if self.description:
            matcher = hc.described_as(self.description, matcher)

        return matcher

    def _find_matcher(self, alias):
        """ Finds a matcher based on the given alias or raises an error if no
            matcher could be found.
        """
        matcher = lookup(alias)
        if not matcher:
            msg = 'Matcher "%s" not found' % alias

            # Try to find similarly named matchers to help the user
            similar = suggest(alias, max=3, cutoff=0.5)
            if len(similar) > 1:
                last = similar.pop()
                msg += '. Perhaps you meant to use %s or %s?' % (', '.join(similar), last)
            elif len(similar) > 0:
                msg += '. Perhaps you meant to use %s?' % similar.pop()

            raise KeyError(msg)

        return matcher

    def _init_matcher(self, *args, **kwargs):
        """ Executes the current matcher appending it to the expression """

        # If subject-less expectation are provided as arguments convert them
        # to plain Hamcrest matchers in order to allow complex compositions
        fn = lambda x: x.evaluate() if isinstance(x, Expectation) else x
        args = [fn(x) for x in args]
        kwargs = dict((k, fn(v)) for k, v in kwargs.items())

        matcher = self.matcher(*args, **kwargs)
        self.expr.append(matcher)
        self.matcher = None
        return matcher

    def described_as(self, description, *args):
        """ Specify a custom message for the matcher """
        if len(args):
            description = description.format(*args)
        self.description = description
        return self

    def desc(self, description, *args):
        """ Just an alias to described_as """
        return self.described_as(description, *args)

    def __getattribute__(self, name):
        """ Hijack property access to handle some special cases.
            Since we might have patched the root object to include
            the `should` properties, we have to capture their use
            here.
        """

        # Ignore .should. style properties
        lowname = name.lower()
        if lowname in ('should', 'to'):
            return self
        if lowname == 'should_not':
            return ExpectationNot(
                self.value,
                self.deferred,
                self.description,
                self.factory,
                self.def_op,
                self.def_matcher
            )

        return object.__getattribute__(self, name)

    def __getattr__(self, name):
        """ Overload property access to interpret them as matchers. """

        # Ignore private (protocol) methods
        if name[0:2] == '__':
            raise AttributeError

        # In factory mode we always create a new instance. This avoids
        # problems when defining multiple expectations using the `should`
        # keyword without resolving every expectation in order.
        obj = self.clone() if self.factory else self

        # If we still have an uninitialized matcher then init it now
        if obj.matcher:
            obj._init_matcher()
            # In deferred mode we will resolve in the __ror__ overload
            if not obj.deferred:
                obj.resolve(obj.value)

        # Normalize the name
        name = re.sub(r'([a-z])([A-Z])', r'\1_\2', name)
        parts = name.lower().split('_')

        # Check if we have a coordinator as first item
        expr = []
        if parts[0] == 'and':
            expr.append(OPERATOR.AND)
            parts.pop(0)
        elif parts[0] == 'or':
            expr.append(OPERATOR.OR)
            parts.pop(0)
        elif parts[0] == 'but':
            expr.append(OPERATOR.BUT)
            parts.pop(0)
        # If no coordinator is given assume a default one
        elif len(obj.expr):
            expr.append(obj.def_op)

        # Negation can come just after a combinator (ie: .and_not_be_equal)
        if 'not' in parts:
            expr.append(OPERATOR.NOT)
            parts.pop(parts.index('not'))

        if len(parts):
            name = '_'.join(parts)
        else:
            name = obj.last_matcher or obj.def_matcher

        # Find a matcher for the computed name
        try:
            obj.matcher = obj._find_matcher(name)
            obj.last_matcher = name
            obj.expr.extend(expr)
        except KeyError as ex:
            # Signal correctly for `hasattr`
            raise AttributeError(str(ex))

        return obj

    def __call__(self, *args, **kwargs):
        """ Execute the matcher just registered by __getattr__ passing any given
            arguments. If we're in deferred mode we don't resolve the matcher yet,
            it'll be done in the __ror__ overload.
        """
        # When called directly (ie: should(foo).xxx) register the param as a transform
        if (len(args) == 1 and hasattr(args[0], '__call__')
                and not self.expr and not self.matcher):
            # We have to clone the expectation so we play fair with the `should` shortcut
            clone = self.clone()
            clone.transform = args[0]
            return clone

        if not self.matcher:
            raise TypeError('No matchers set. Usage: <value> | should.<matcher>(<expectation>)')

        self._init_matcher(*args, **kwargs)

        # In deferred mode we will resolve in the __ror__ overload
        if not self.deferred:
            self.resolve(self.value)

        return self

    def __enter__(self):
        """ Implements the context manager protocol. Specially useful for asserting exceptions
        """
        clone = self.clone()
        self._contexts.append(clone)
        self.reset()
        return self

    def __exit__(self, exc, value, trace):
        # If an assertion failed inside the block just raise that one
        if isinstance(value, AssertionError):
            return False

        expr = self._contexts.pop()
        result = ContextManagerResult(exc, value, trace)
        expr.resolve(result)
        return True

    def __eq__(self, other):
        """ Overloads the equality operator to trigger a resolution of the matcher
            against the other expression value. This allows to easily use expressions
            in other libraries like Mock.
        """
        clone = self.clone()
        try:
            clone.resolve(other)
            return True
        except AssertionError:
            return False
        # Any exception is silenced and we just return false
        except:
            return False

    def __ne__(self, other):
        """ Overload not equal since Python will default to identity instead of negating
            the result from equality.
        """
        return not self.__eq__(other)

    def __repr__(self):
        """ This is specially useful when using the library on an interactive interpreter
        """
        exp = self.clone()
        if exp.matcher:
            exp._init_matcher()

        if not exp.expr:
            return 'Uninitialized expectation <{0}>'.format(self.__class__.__name__)

        matcher = exp.evaluate()
        return str(matcher)


class ExpectationNot(Expectation):
    """ Negates the result of the matcher """

    def _assertion(self, matcher, value):
        hc.assert_that(value, IsNot(matcher))


class ExpectationAny(Expectation):
    """ Succeeds if any of the items in an iterable value passes the matcher """

    def _assertion(self, matcher, value):
        hc.assert_that(value, hc.has_item(matcher))

    def _transform(self, value):
        if self.transform:
            value = [super(ExpectationAny, self)._transform(x) for x in value]
        return value


class ExpectationAll(ExpectationAny):
    """ Succeeds if all of the items in an iterable value pass the matcher """

    def _assertion(self, matcher, value):
        hc.assert_that(value, hc.only_contains(matcher))


class ExpectationNone(ExpectationAny):
    """ Succeeds if none of the items in an iterable value passes the matcher """

    def _assertion(self, matcher, value):
        hc.assert_that(value, IsNot(hc.has_item(matcher)))



"""
Patches mockito's param matcher to allow pyshould expectations.

NOTE: This module has been deprecated, should expectations can
      now directly work with third party libraries by the
      overloading of the equality operators.
"""

__author__ = "Ivan -DrSlump- Montes"
__email__ = "drslump@pollinimini.net"
__license__ = "MIT"

from warnings import warn

warn("patch_mockito is no longer needed", DeprecationWarning)
